@@ -3,23 +3,42 @@ from itertools import islice
 
 import numpy as np
 
+from zarr._storage.store import (_get_metadata_suffix, data_root, meta_root,
+                                 DEFAULT_ZARR_VERSION, assert_zarr_v3_api_available)
 from zarr.attrs import Attributes
 from zarr.core import Array
 from zarr.creation import (array, create, empty, empty_like, full, full_like,
-                           normalize_store_arg, ones, ones_like, zeros,
-                           zeros_like)
+                           ones, ones_like, zeros, zeros_like)
 from zarr.errors import (
     ContainsArrayError,
     ContainsGroupError,
     GroupNotFoundError,
     ReadOnlyError,
 )
-from zarr.meta import decode_group_metadata
-from zarr.storage import (MemoryStore, attrs_key, contains_array,
-                          contains_group, group_meta_key, init_group, listdir,
-                          rename, rmdir)
-from zarr.util import (InfoReporter, TreeViewer, is_valid_python_name, nolock,
-                       normalize_shape, normalize_storage_path)
+from zarr.storage import (
+    _get_hierarchy_metadata,
+    _prefix_to_group_key,
+    BaseStore,
+    MemoryStore,
+    attrs_key,
+    contains_array,
+    contains_group,
+    group_meta_key,
+    init_group,
+    listdir,
+    normalize_store_arg,
+    rename,
+    rmdir,
+)
+from zarr._storage.v3 import MemoryStoreV3
+from zarr.util import (
+    InfoReporter,
+    TreeViewer,
+    is_valid_python_name,
+    nolock,
+    normalize_shape,
+    normalize_storage_path,
+)
 
 
 class Group(MutableMapping):
@@ -45,6 +64,12 @@ class Group(MutableMapping):
     synchronizer : object, optional
         Array synchronizer.
 
+    meta_array : array-like, optional
+        An array instance to use for determining arrays to create and return
+        to users. Use `numpy.empty(())` by default.
+
+        .. versionadded:: 2.13
+
     Attributes
     ----------
     store
@@ -55,6 +80,7 @@ class Group(MutableMapping):
     synchronizer
     attrs
     info
+    meta_array
 
     Methods
     -------
@@ -95,7 +121,17 @@ class Group(MutableMapping):
     """
 
     def __init__(self, store, path=None, read_only=False, chunk_store=None,
-                 cache_attrs=True, synchronizer=None):
+                 cache_attrs=True, synchronizer=None, zarr_version=None, *,
+                 meta_array=None):
+        store: BaseStore = _normalize_store_arg(store, zarr_version=zarr_version)
+        if zarr_version is None:
+            zarr_version = getattr(store, '_store_version', DEFAULT_ZARR_VERSION)
+
+        if zarr_version != 2:
+            assert_zarr_v3_api_available()
+
+        if chunk_store is not None:
+            chunk_store: BaseStore = _normalize_store_arg(chunk_store, zarr_version=zarr_version)
         self._store = store
         self._chunk_store = chunk_store
         self._path = normalize_storage_path(path)
@@ -105,23 +141,47 @@ class Group(MutableMapping):
             self._key_prefix = ''
         self._read_only = read_only
         self._synchronizer = synchronizer
+        if meta_array is not None:
+            self._meta_array = np.empty_like(meta_array, shape=())
+        else:
+            self._meta_array = np.empty(())
+        self._version = zarr_version
+        if self._version == 3:
+            self._data_key_prefix = data_root + self._key_prefix
+            self._data_path = data_root + self._path
+            self._hierarchy_metadata = _get_hierarchy_metadata(store=self._store)
+            self._metadata_key_suffix = _get_metadata_suffix(store=self._store)
 
         # guard conditions
         if contains_array(store, path=self._path):
             raise ContainsArrayError(path)
 
         # initialize metadata
+        mkey = None
         try:
-            mkey = self._key_prefix + group_meta_key
+            mkey = _prefix_to_group_key(self._store, self._key_prefix)
+            assert not mkey.endswith("root/.group")
             meta_bytes = store[mkey]
         except KeyError:
-            raise GroupNotFoundError(path)
+            if self._version == 2:
+                raise GroupNotFoundError(path)
+            else:
+                implicit_prefix = meta_root + self._key_prefix
+                if self._store.list_prefix(implicit_prefix):
+                    # implicit group does not have any metadata
+                    self._meta = None
+                else:
+                    raise GroupNotFoundError(path)
         else:
-            meta = decode_group_metadata(meta_bytes)
-            self._meta = meta
+            self._meta = self._store._metadata_class.decode_group_metadata(meta_bytes)
 
         # setup attributes
-        akey = self._key_prefix + attrs_key
+        if self._version == 2:
+            akey = self._key_prefix + attrs_key
+        else:
+            # Note: mkey doesn't actually exist for implicit groups, but the
+            # object can still be created.
+            akey = mkey
         self._attrs = Attributes(store, key=akey, read_only=read_only,
                                  cache=cache_attrs, synchronizer=synchronizer)
 
@@ -183,6 +243,13 @@ class Group(MutableMapping):
         """Return diagnostic information about the group."""
         return self._info
 
+    @property
+    def meta_array(self):
+        """An array-like instance to use for determining arrays to create and return
+        to users.
+        """
+        return self._meta_array
+
     def __eq__(self, other):
         return (
             isinstance(other, Group) and
@@ -212,11 +279,36 @@ class Group(MutableMapping):
         quux
 
         """
-        for key in sorted(listdir(self._store, self._path)):
-            path = self._key_prefix + key
-            if (contains_array(self._store, path) or
-                    contains_group(self._store, path)):
-                yield key
+        if getattr(self._store, '_store_version', 2) == 2:
+            for key in sorted(listdir(self._store, self._path)):
+                path = self._key_prefix + key
+                if (contains_array(self._store, path) or
+                        contains_group(self._store, path)):
+                    yield key
+        else:
+            # TODO: Should this iterate over data folders and/or metadata
+            #       folders and/or metadata files
+
+            dir_path = meta_root + self._key_prefix
+            name_start = len(dir_path)
+            keys, prefixes = self._store.list_dir(dir_path)
+
+            # yield any groups or arrays
+            sfx = self._metadata_key_suffix
+            for key in keys:
+                len_suffix = len('.group') + len(sfx)  # same for .array
+                if key.endswith(('.group' + sfx, '.array' + sfx)):
+                    yield key[name_start:-len_suffix]
+
+            # also yield any implicit groups
+            for prefix in prefixes:
+                prefix = prefix.rstrip('/')
+                # only implicit if there is no .group.sfx file
+                if not prefix + '.group' + sfx in self._store:
+                    yield prefix[name_start:]
+
+            # Note: omit data/root/ to avoid duplicate listings
+            #       any group in data/root/ must has an entry in meta/root/
 
     def __len__(self):
         """Number of members."""
@@ -237,11 +329,8 @@ class Group(MutableMapping):
         return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
-        """If the underlying Store has a ``close`` method, call it."""
-        try:
-            self.store.close()
-        except AttributeError:
-            pass
+        """Call the close method of the underlying Store."""
+        self.store.close()
 
     def info_items(self):
 
@@ -281,11 +370,19 @@ class Group(MutableMapping):
         return items
 
     def __getstate__(self):
-        return (self._store, self._path, self._read_only, self._chunk_store,
-                self.attrs.cache, self._synchronizer)
+        return {
+            "store": self._store,
+            "path": self._path,
+            "read_only": self._read_only,
+            "chunk_store": self._chunk_store,
+            "cache_attrs": self._attrs.cache,
+            "synchronizer": self._synchronizer,
+            "zarr_version": self._version,
+            "meta_array": self._meta_array,
+        }
 
     def __setstate__(self, state):
-        self.__init__(*state)
+        self.__init__(**state)
 
     def _item_path(self, item):
         absolute = isinstance(item, str) and item and item[0] == '/'
@@ -313,7 +410,7 @@ class Group(MutableMapping):
         """
         path = self._item_path(item)
         return contains_array(self._store, path) or \
-            contains_group(self._store, path)
+            contains_group(self._store, path, explicit_only=False)
 
     def __getitem__(self, item):
         """Obtain a group member.
@@ -340,11 +437,23 @@ class Group(MutableMapping):
         if contains_array(self._store, path):
             return Array(self._store, read_only=self._read_only, path=path,
                          chunk_store=self._chunk_store,
-                         synchronizer=self._synchronizer, cache_attrs=self.attrs.cache)
-        elif contains_group(self._store, path):
+                         synchronizer=self._synchronizer, cache_attrs=self.attrs.cache,
+                         zarr_version=self._version, meta_array=self._meta_array)
+        elif contains_group(self._store, path, explicit_only=True):
             return Group(self._store, read_only=self._read_only, path=path,
                          chunk_store=self._chunk_store, cache_attrs=self.attrs.cache,
-                         synchronizer=self._synchronizer)
+                         synchronizer=self._synchronizer, zarr_version=self._version,
+                         meta_array=self._meta_array)
+        elif self._version == 3:
+            implicit_group = meta_root + path + '/'
+            # non-empty folder in the metadata path implies an implicit group
+            if self._store.list_prefix(implicit_group):
+                return Group(self._store, read_only=self._read_only, path=path,
+                             chunk_store=self._chunk_store, cache_attrs=self.attrs.cache,
+                             synchronizer=self._synchronizer, zarr_version=self._version,
+                             meta_array=self._meta_array)
+            else:
+                raise KeyError(item)
         else:
             raise KeyError(item)
 
@@ -357,7 +466,7 @@ class Group(MutableMapping):
     def _delitem_nosync(self, item):
         path = self._item_path(item)
         if contains_array(self._store, path) or \
-                contains_group(self._store, path):
+                contains_group(self._store, path, explicit_only=False):
             rmdir(self._store, path)
         else:
             raise KeyError(item)
@@ -394,10 +503,29 @@ class Group(MutableMapping):
         ['bar', 'foo']
 
         """
-        for key in sorted(listdir(self._store, self._path)):
-            path = self._key_prefix + key
-            if contains_group(self._store, path):
-                yield key
+        if self._version == 2:
+            for key in sorted(listdir(self._store, self._path)):
+                path = self._key_prefix + key
+                if contains_group(self._store, path):
+                    yield key
+        else:
+            dir_name = meta_root + self._path
+            group_sfx = '.group' + self._metadata_key_suffix
+            # The fact that we call sorted means this can't be a streaming generator.
+            # The keys are already in memory.
+            all_keys = sorted(listdir(self._store, dir_name))
+            for key in all_keys:
+                if key.endswith(group_sfx):
+                    key = key[:-len(group_sfx)]
+                    if key in all_keys:
+                        # otherwise we will double count this group
+                        continue
+                path = self._key_prefix + key
+                if path.endswith(".array" + self._metadata_key_suffix):
+                    # skip array keys
+                    continue
+                if contains_group(self._store, path, explicit_only=False):
+                    yield key
 
     def groups(self):
         """Return an iterator over (name, value) pairs for groups only.
@@ -416,13 +544,30 @@ class Group(MutableMapping):
         foo <class 'zarr.hierarchy.Group'>
 
         """
-        for key in sorted(listdir(self._store, self._path)):
-            path = self._key_prefix + key
-            if contains_group(self._store, path):
-                yield key, Group(self._store, path=path, read_only=self._read_only,
-                                 chunk_store=self._chunk_store,
-                                 cache_attrs=self.attrs.cache,
-                                 synchronizer=self._synchronizer)
+        if self._version == 2:
+            for key in sorted(listdir(self._store, self._path)):
+                path = self._key_prefix + key
+                if contains_group(self._store, path, explicit_only=False):
+                    yield key, Group(
+                        self._store,
+                        path=path,
+                        read_only=self._read_only,
+                        chunk_store=self._chunk_store,
+                        cache_attrs=self.attrs.cache,
+                        synchronizer=self._synchronizer,
+                        zarr_version=self._version)
+
+        else:
+            for key in self.group_keys():
+                path = self._key_prefix + key
+                yield key, Group(
+                    self._store,
+                    path=path,
+                    read_only=self._read_only,
+                    chunk_store=self._chunk_store,
+                    cache_attrs=self.attrs.cache,
+                    synchronizer=self._synchronizer,
+                    zarr_version=self._version)
 
     def array_keys(self, recurse=False):
         """Return an iterator over member names for arrays only.
@@ -479,14 +624,34 @@ class Group(MutableMapping):
                                 recurse=recurse)
 
     def _array_iter(self, keys_only, method, recurse):
-        for key in sorted(listdir(self._store, self._path)):
-            path = self._key_prefix + key
-            if contains_array(self._store, path):
-                yield key if keys_only else (key, self[key])
-            elif recurse and contains_group(self._store, path):
-                group = self[key]
-                for i in getattr(group, method)(recurse=recurse):
-                    yield i
+        if self._version == 2:
+            for key in sorted(listdir(self._store, self._path)):
+                path = self._key_prefix + key
+                if contains_array(self._store, path):
+                    _key = key.rstrip("/")
+                    yield _key if keys_only else (_key, self[key])
+                elif recurse and contains_group(self._store, path):
+                    group = self[key]
+                    yield from getattr(group, method)(recurse=recurse)
+        else:
+            dir_name = meta_root + self._path
+            array_sfx = '.array' + self._metadata_key_suffix
+            group_sfx = '.group' + self._metadata_key_suffix
+
+            for key in sorted(listdir(self._store, dir_name)):
+                if key.endswith(array_sfx):
+                    key = key[:-len(array_sfx)]
+                    _key = key.rstrip("/")
+                    yield _key if keys_only else (_key, self[key])
+
+                path = self._key_prefix + key
+                assert not path.startswith("meta/")
+                if key.endswith(group_sfx):
+                    # skip group metadata keys
+                    continue
+                elif recurse and contains_group(self._store, path):
+                    group = self[key]
+                    yield from getattr(group, method)(recurse=recurse)
 
     def visitvalues(self, func):
         """Run ``func`` on each object.
@@ -520,8 +685,7 @@ class Group(MutableMapping):
             yield obj
             keys = sorted(getattr(obj, "keys", lambda: [])())
             for k in keys:
-                for v in _visit(obj[k]):
-                    yield v
+                yield from _visit(obj[k])
 
         for each_obj in islice(_visit(self), 1, None):
             value = func(each_obj)
@@ -554,6 +718,76 @@ class Group(MutableMapping):
         baz
         quux
 
+        Search for members matching some name query can be implemented using
+        ``visit`` that is, ``find`` and ``findall``. Consider the following
+        tree::
+
+            /
+            ├── aaa
+            │   └── bbb
+            │       └── ccc
+            │           └── aaa
+            ├── bar
+            └── foo
+
+        It is created as follows:
+
+        >>> root = zarr.group()
+        >>> foo = root.create_group("foo")
+        >>> bar = root.create_group("bar")
+        >>> root.create_group("aaa").create_group("bbb").create_group("ccc").create_group("aaa")
+        <zarr.hierarchy.Group '/aaa/bbb/ccc/aaa'>
+
+        For ``find``, the first path that matches a given pattern (for example
+        "aaa") is returned. Note that a non-None value is returned in the visit
+        function to stop further iteration.
+
+        >>> import re
+        >>> pattern = re.compile("aaa")
+        >>> found = None
+        >>> def find(path):
+        ...     global found
+        ...     if pattern.search(path) is not None:
+        ...         found = path
+        ...         return True
+        ...
+        >>> root.visit(find)
+        True
+        >>> print(found)
+        aaa
+
+        For ``findall``, all the results are gathered into a list
+
+        >>> pattern = re.compile("aaa")
+        >>> found = []
+        >>> def findall(path):
+        ...     if pattern.search(path) is not None:
+        ...         found.append(path)
+        ...
+        >>> root.visit(findall)
+        >>> print(found)
+        ['aaa', 'aaa/bbb', 'aaa/bbb/ccc', 'aaa/bbb/ccc/aaa']
+
+        To match only on the last part of the path, use a greedy regex to filter
+        out the prefix:
+
+        >>> prefix_pattern = re.compile(r".*/")
+        >>> pattern = re.compile("aaa")
+        >>> found = []
+        >>> def findall(path):
+        ...     match = prefix_pattern.match(path)
+        ...     if match is None:
+        ...         name = path
+        ...     else:
+        ...         _, end = match.span()
+        ...         name = path[end:]
+        ...     if pattern.search(name) is not None:
+        ...         found.append(path)
+        ...     return None
+        ...
+        >>> root.visit(findall)
+        >>> print(found)
+        ['aaa', 'aaa/bbb/ccc/aaa']
         """
 
         base_len = len(self.name)
@@ -695,7 +929,7 @@ class Group(MutableMapping):
 
         return Group(self._store, path=path, read_only=self._read_only,
                      chunk_store=self._chunk_store, cache_attrs=self.attrs.cache,
-                     synchronizer=self._synchronizer)
+                     synchronizer=self._synchronizer, zarr_version=self._version)
 
     def create_groups(self, *names, **kwargs):
         """Convenience method to create multiple groups in a single call."""
@@ -739,7 +973,7 @@ class Group(MutableMapping):
 
         return Group(self._store, path=path, read_only=self._read_only,
                      chunk_store=self._chunk_store, cache_attrs=self.attrs.cache,
-                     synchronizer=self._synchronizer)
+                     synchronizer=self._synchronizer, zarr_version=self._version)
 
     def require_groups(self, *names):
         """Convenience method to require multiple groups in a single call."""
@@ -756,7 +990,7 @@ class Group(MutableMapping):
         ----------
         name : string
             Array name.
-        data : array_like, optional
+        data : array-like, optional
             Initial data.
         shape : int or tuple of ints
             Array shape.
@@ -804,11 +1038,13 @@ class Group(MutableMapping):
         <zarr.core.Array '/bar/baz/qux' (100, 100, 100) float64>
 
         """
+        assert "mode" not in kwargs
 
         return self._write_op(self._create_dataset_nosync, name, **kwargs)
 
     def _create_dataset_nosync(self, name, data=None, **kwargs):
 
+        assert "mode" not in kwargs
         path = self._item_path(name)
 
         # determine synchronizer
@@ -865,7 +1101,8 @@ class Group(MutableMapping):
             cache_attrs = kwargs.get('cache_attrs', self.attrs.cache)
             a = Array(self._store, path=path, read_only=self._read_only,
                       chunk_store=self._chunk_store, synchronizer=synchronizer,
-                      cache_metadata=cache_metadata, cache_attrs=cache_attrs)
+                      cache_metadata=cache_metadata, cache_attrs=cache_attrs,
+                      meta_array=self._meta_array)
             shape = normalize_shape(shape)
             if shape != a.shape:
                 raise TypeError('shape do not match existing array; expected {}, got {}'
@@ -1025,9 +1262,10 @@ class Group(MutableMapping):
 
         # Check that source exists.
         if not (contains_array(self._store, source) or
-                contains_group(self._store, source)):
+                contains_group(self._store, source, explicit_only=False)):
             raise ValueError('The source, "%s", does not exist.' % source)
-        if contains_array(self._store, dest) or contains_group(self._store, dest):
+        if (contains_array(self._store, dest) or
+                contains_group(self._store, dest, explicit_only=False)):
             raise ValueError('The dest, "%s", already exists.' % dest)
 
         # Ensure groups needed for `dest` exist.
@@ -1037,15 +1275,23 @@ class Group(MutableMapping):
         self._write_op(self._move_nosync, source, dest)
 
 
-def _normalize_store_arg(store, *, clobber=False, storage_options=None, mode=None):
+def _normalize_store_arg(store, *, storage_options=None, mode="r",
+                         zarr_version=None):
+    if zarr_version is None:
+        zarr_version = getattr(store, '_store_version', DEFAULT_ZARR_VERSION)
+
+    if zarr_version != 2:
+        assert_zarr_v3_api_available()
+
     if store is None:
-        return MemoryStore()
-    return normalize_store_arg(store, clobber=clobber,
-                               storage_options=storage_options, mode=mode)
+        return MemoryStore() if zarr_version == 2 else MemoryStoreV3()
+    return normalize_store_arg(store,
+                               storage_options=storage_options, mode=mode,
+                               zarr_version=zarr_version)
 
 
 def group(store=None, overwrite=False, chunk_store=None,
-          cache_attrs=True, synchronizer=None, path=None):
+          cache_attrs=True, synchronizer=None, path=None, *, zarr_version=None):
     """Create a group.
 
     Parameters
@@ -1090,20 +1336,32 @@ def group(store=None, overwrite=False, chunk_store=None,
     """
 
     # handle polymorphic store arg
-    store = _normalize_store_arg(store)
+    store = _normalize_store_arg(store, zarr_version=zarr_version)
+    if zarr_version is None:
+        zarr_version = getattr(store, '_store_version', DEFAULT_ZARR_VERSION)
+
+    if zarr_version != 2:
+        assert_zarr_v3_api_available()
+
     path = normalize_storage_path(path)
 
-    # require group
-    if overwrite or not contains_group(store):
+    requires_init = None
+    if zarr_version == 2:
+        requires_init = overwrite or not contains_group(store)
+    elif zarr_version == 3:
+        requires_init = overwrite or not contains_group(store, path)
+
+    if requires_init:
         init_group(store, overwrite=overwrite, chunk_store=chunk_store,
                    path=path)
 
     return Group(store, read_only=False, chunk_store=chunk_store,
-                 cache_attrs=cache_attrs, synchronizer=synchronizer, path=path)
+                 cache_attrs=cache_attrs, synchronizer=synchronizer, path=path,
+                 zarr_version=zarr_version)
 
 
 def open_group(store=None, mode='a', cache_attrs=True, synchronizer=None, path=None,
-               chunk_store=None, storage_options=None):
+               chunk_store=None, storage_options=None, *, zarr_version=None, meta_array=None):
     """Open a group using file-mode-like semantics.
 
     Parameters
@@ -1128,6 +1386,11 @@ def open_group(store=None, mode='a', cache_attrs=True, synchronizer=None, path=N
     storage_options : dict
         If using an fsspec URL to create the store, these will be passed to
         the backend implementation. Ignored otherwise.
+    meta_array : array-like, optional
+        An array instance to use for determining arrays to create and return
+        to users. Use `numpy.empty(())` by default.
+
+        .. versionadded:: 2.13
 
     Returns
     -------
@@ -1150,13 +1413,25 @@ def open_group(store=None, mode='a', cache_attrs=True, synchronizer=None, path=N
     """
 
     # handle polymorphic store arg
-    clobber = mode != "r"
     store = _normalize_store_arg(
-        store, clobber=clobber, storage_options=storage_options, mode=mode
-    )
+        store, storage_options=storage_options, mode=mode,
+        zarr_version=zarr_version)
+    if zarr_version is None:
+        zarr_version = getattr(store, '_store_version', DEFAULT_ZARR_VERSION)
+
+    if zarr_version != 2:
+        assert_zarr_v3_api_available()
+
     if chunk_store is not None:
-        chunk_store = _normalize_store_arg(chunk_store, clobber=clobber,
-                                           storage_options=storage_options)
+        chunk_store = _normalize_store_arg(chunk_store,
+                                           storage_options=storage_options,
+                                           mode=mode,
+                                           zarr_version=zarr_version)
+        if getattr(chunk_store, '_store_version', DEFAULT_ZARR_VERSION) != zarr_version:
+            raise ValueError(  # pragma: no cover
+                "zarr_version of store and chunk_store must match"
+            )
+
     path = normalize_storage_path(path)
 
     # ensure store is initialized
@@ -1188,4 +1463,5 @@ def open_group(store=None, mode='a', cache_attrs=True, synchronizer=None, path=N
     read_only = mode == 'r'
 
     return Group(store, read_only=read_only, cache_attrs=cache_attrs,
-                 synchronizer=synchronizer, path=path, chunk_store=chunk_store)
+                 synchronizer=synchronizer, path=path, chunk_store=chunk_store,
+                 zarr_version=zarr_version, meta_array=meta_array)
