@@ -3,6 +3,7 @@
 import os
 import struct
 import sys
+from typing import Any, Dict, Optional, cast
 import warnings
 
 import numpy as np
@@ -17,6 +18,8 @@ from .storage import array_meta_key as zarr_array_meta_key
 from .storage import attrs_key as zarr_attrs_key
 from .storage import group_meta_key as zarr_group_meta_key
 
+N5_FORMAT = '2.0.0'
+
 zarr_to_n5_keys = [
     ('chunks', 'blockSize'),
     ('dtype', 'dataType'),
@@ -38,7 +41,7 @@ class N5Store(NestedDirectoryStore):
     normalize_keys : bool, optional
         If True, all store keys will be normalized to use lower case characters
         (e.g. 'foo' and 'FOO' will be treated as equivalent). This can be
-        useful to avoid potential discrepancies between case-senstive and
+        useful to avoid potential discrepancies between case-sensitive and
         case-insensitive file system. Default value is False.
 
     Examples
@@ -67,70 +70,70 @@ class N5Store(NestedDirectoryStore):
 
     """
 
-    def __getitem__(self, key):
-
+    def __getitem__(self, key: str) -> bytes:
         if key.endswith(zarr_group_meta_key):
 
-            key = key.replace(zarr_group_meta_key, n5_attrs_key)
-            value = group_metadata_to_zarr(self._load_n5_attrs(key))
+            key_new = key.replace(zarr_group_meta_key, n5_attrs_key)
+            value = group_metadata_to_zarr(self._load_n5_attrs(key_new))
 
             return json_dumps(value)
 
         elif key.endswith(zarr_array_meta_key):
 
-            key = key.replace(zarr_array_meta_key, n5_attrs_key)
-            value = array_metadata_to_zarr(self._load_n5_attrs(key))
-
+            key_new = key.replace(zarr_array_meta_key, n5_attrs_key)
+            top_level = key == zarr_array_meta_key
+            value = array_metadata_to_zarr(self._load_n5_attrs(key_new), top_level=top_level)
             return json_dumps(value)
 
         elif key.endswith(zarr_attrs_key):
 
-            key = key.replace(zarr_attrs_key, n5_attrs_key)
-            value = attrs_to_zarr(self._load_n5_attrs(key))
+            key_new = key.replace(zarr_attrs_key, n5_attrs_key)
+            value = attrs_to_zarr(self._load_n5_attrs(key_new))
 
             if len(value) == 0:
-                raise KeyError(key)
+                raise KeyError(key_new)
             else:
                 return json_dumps(value)
 
         elif is_chunk_key(key):
+            key_new = invert_chunk_coords(key)
 
-            key = invert_chunk_coords(key)
+        else:
+            key_new = key
 
-        return super().__getitem__(key)
+        return super().__getitem__(key_new)
 
-    def __setitem__(self, key, value):
+    def __setitem__(self, key: str, value: Any):
 
         if key.endswith(zarr_group_meta_key):
 
-            key = key.replace(zarr_group_meta_key, n5_attrs_key)
+            key_new = key.replace(zarr_group_meta_key, n5_attrs_key)
 
-            n5_attrs = self._load_n5_attrs(key)
+            n5_attrs = self._load_n5_attrs(key_new)
             n5_attrs.update(**group_metadata_to_n5(json_loads(value)))
 
             value = json_dumps(n5_attrs)
 
         elif key.endswith(zarr_array_meta_key):
 
-            key = key.replace(zarr_array_meta_key, n5_attrs_key)
-
-            n5_attrs = self._load_n5_attrs(key)
-            n5_attrs.update(**array_metadata_to_n5(json_loads(value)))
-
+            key_new = key.replace(zarr_array_meta_key, n5_attrs_key)
+            top_level = key == zarr_array_meta_key
+            n5_attrs = self._load_n5_attrs(key_new)
+            n5_attrs.update(**array_metadata_to_n5(json_loads(value), top_level=top_level))
             value = json_dumps(n5_attrs)
 
         elif key.endswith(zarr_attrs_key):
 
-            key = key.replace(zarr_attrs_key, n5_attrs_key)
+            key_new = key.replace(zarr_attrs_key, n5_attrs_key)
 
-            n5_attrs = self._load_n5_attrs(key)
+            n5_attrs = self._load_n5_attrs(key_new)
             zarr_attrs = json_loads(value)
 
             for k in n5_keywords:
-                if k in zarr_attrs.keys():
-                    warnings.warn("attribute %s is a reserved N5 keyword" % k, UserWarning)
+                if k in zarr_attrs:
+                    warnings.warn(f"Attribute {k} is a reserved N5 keyword", UserWarning)
 
-            # replace previous user attributes
+            # remove previous user attributes
             for k in list(n5_attrs.keys()):
                 if k not in n5_keywords:
                     del n5_attrs[k]
@@ -141,50 +144,55 @@ class N5Store(NestedDirectoryStore):
             value = json_dumps(n5_attrs)
 
         elif is_chunk_key(key):
+            key_new = invert_chunk_coords(key)
 
-            key = invert_chunk_coords(key)
+        else:
+            key_new = key
 
-        super().__setitem__(key, value)
+        super().__setitem__(key_new, value)
 
-    def __delitem__(self, key):
-
-        if key.endswith(zarr_group_meta_key):  # pragma: no cover
-            key = key.replace(zarr_group_meta_key, n5_attrs_key)
-        elif key.endswith(zarr_array_meta_key):  # pragma: no cover
-            key = key.replace(zarr_array_meta_key, n5_attrs_key)
-        elif key.endswith(zarr_attrs_key):  # pragma: no cover
-            key = key.replace(zarr_attrs_key, n5_attrs_key)
+    def __delitem__(self, key: str):
+        if key.endswith(zarr_group_meta_key):
+            key_new = key.replace(zarr_group_meta_key, n5_attrs_key)
+        elif key.endswith(zarr_array_meta_key):
+            key_new = key.replace(zarr_array_meta_key, n5_attrs_key)
+        elif key.endswith(zarr_attrs_key):
+            key_new = key.replace(zarr_attrs_key, n5_attrs_key)
         elif is_chunk_key(key):
-            key = invert_chunk_coords(key)
+            key_new = invert_chunk_coords(key)
+        else:
+            key_new = key
 
-        super().__delitem__(key)
+        super().__delitem__(key_new)
 
     def __contains__(self, key):
 
         if key.endswith(zarr_group_meta_key):
 
-            key = key.replace(zarr_group_meta_key, n5_attrs_key)
-            if key not in self:
+            key_new = key.replace(zarr_group_meta_key, n5_attrs_key)
+            if key_new not in self:
                 return False
             # group if not a dataset (attributes do not contain 'dimensions')
-            return 'dimensions' not in self._load_n5_attrs(key)
+            return 'dimensions' not in self._load_n5_attrs(key_new)
 
         elif key.endswith(zarr_array_meta_key):
 
-            key = key.replace(zarr_array_meta_key, n5_attrs_key)
+            key_new = key.replace(zarr_array_meta_key, n5_attrs_key)
             # array if attributes contain 'dimensions'
-            return 'dimensions' in self._load_n5_attrs(key)
+            return 'dimensions' in self._load_n5_attrs(key_new)
 
         elif key.endswith(zarr_attrs_key):
 
-            key = key.replace(zarr_attrs_key, n5_attrs_key)
-            return self._contains_attrs(key)
+            key_new = key.replace(zarr_attrs_key, n5_attrs_key)
+            return self._contains_attrs(key_new)
 
         elif is_chunk_key(key):
 
-            key = invert_chunk_coords(key)
+            key_new = invert_chunk_coords(key)
+        else:
+            key_new = key
 
-        return super().__contains__(key)
+        return super().__contains__(key_new)
 
     def __eq__(self, other):
         return (
@@ -192,11 +200,11 @@ class N5Store(NestedDirectoryStore):
             self.path == other.path
         )
 
-    def listdir(self, path=None):
+    def listdir(self, path: Optional[str] = None):
 
         if path is not None:
             path = invert_chunk_coords(path)
-
+        path = cast(str, path)
         # We can't use NestedDirectoryStore's listdir, as it requires
         # array_meta_key to be present in array directories, which this store
         # doesn't provide.
@@ -233,7 +241,7 @@ class N5Store(NestedDirectoryStore):
             # replace n5 attribute file with respective zarr attribute files
             children.remove(n5_attrs_key)
             children.append(zarr_group_meta_key)
-            if self._contains_attrs(path):  # pragma: no cover
+            if self._contains_attrs(path):
                 children.append(zarr_attrs_key)
 
             return sorted(children)
@@ -242,14 +250,14 @@ class N5Store(NestedDirectoryStore):
 
             return children
 
-    def _load_n5_attrs(self, path):
+    def _load_n5_attrs(self, path: str) -> Dict[str, Any]:
         try:
             s = super().__getitem__(path)
             return json_loads(s)
         except KeyError:
             return {}
 
-    def _is_group(self, path):
+    def _is_group(self, path: str):
 
         if path is None:
             attrs_key = n5_attrs_key
@@ -259,7 +267,7 @@ class N5Store(NestedDirectoryStore):
         n5_attrs = self._load_n5_attrs(attrs_key)
         return len(n5_attrs) > 0 and 'dimensions' not in n5_attrs
 
-    def _is_array(self, path):
+    def _is_array(self, path: str):
 
         if path is None:
             attrs_key = n5_attrs_key
@@ -268,14 +276,14 @@ class N5Store(NestedDirectoryStore):
 
         return 'dimensions' in self._load_n5_attrs(attrs_key)
 
-    def _contains_attrs(self, path):
+    def _contains_attrs(self, path: str):
 
         if path is None:
             attrs_key = n5_attrs_key
         else:
             if not path.endswith(n5_attrs_key):
                 attrs_key = os.path.join(path, n5_attrs_key)
-            else:  # pragma: no cover
+            else:
                 attrs_key = path
 
         attrs = attrs_to_zarr(self._load_n5_attrs(attrs_key))
@@ -283,8 +291,9 @@ class N5Store(NestedDirectoryStore):
 
 
 class N5FSStore(FSStore):
-    """Implentation of the N5 format (https://github.com/saalfeldlab/n5) using `fsspec`,
-    which allows storage on a variety of filesystems. Based on `zarr.N5Store`.
+    """Implementation of the N5 format (https://github.com/saalfeldlab/n5)
+    using `fsspec`, which allows storage on a variety of filesystems. Based
+    on `zarr.N5Store`.
     Parameters
     ----------
     path : string
@@ -292,7 +301,7 @@ class N5FSStore(FSStore):
     normalize_keys : bool, optional
         If True, all store keys will be normalized to use lower case characters
         (e.g. 'foo' and 'FOO' will be treated as equivalent). This can be
-        useful to avoid potential discrepancies between case-senstive and
+        useful to avoid potential discrepancies between case-sensitive and
         case-insensitive file system. Default value is False.
 
     Examples
@@ -342,7 +351,8 @@ class N5FSStore(FSStore):
         dimension_separator = "."
         super().__init__(*args, dimension_separator=dimension_separator, **kwargs)
 
-    def _swap_separator(self, key):
+    @staticmethod
+    def _swap_separator(key: str):
         segments = list(key.split('/'))
         if segments:
             last_segment = segments[-1]
@@ -353,7 +363,7 @@ class N5FSStore(FSStore):
                 key = '/'.join(segments)
         return key
 
-    def _normalize_key(self, key):
+    def _normalize_key(self, key: str):
         if is_chunk_key(key):
             key = invert_chunk_coords(key)
 
@@ -366,65 +376,68 @@ class N5FSStore(FSStore):
                 key = "/".join(bits + [end])
         return key.lower() if self.normalize_keys else key
 
-    def __getitem__(self, key):
+    def __getitem__(self, key: str) -> bytes:
         if key.endswith(zarr_group_meta_key):
 
-            key = key.replace(zarr_group_meta_key, self._group_meta_key)
-            value = group_metadata_to_zarr(self._load_n5_attrs(key))
+            key_new = key.replace(zarr_group_meta_key, self._group_meta_key)
+            value = group_metadata_to_zarr(self._load_n5_attrs(key_new))
 
             return json_dumps(value)
 
         elif key.endswith(zarr_array_meta_key):
 
-            key = key.replace(zarr_array_meta_key, self._array_meta_key)
-            value = array_metadata_to_zarr(self._load_n5_attrs(key))
-
+            key_new = key.replace(zarr_array_meta_key, self._array_meta_key)
+            top_level = key == zarr_array_meta_key
+            value = array_metadata_to_zarr(self._load_n5_attrs(key_new), top_level=top_level)
             return json_dumps(value)
 
         elif key.endswith(zarr_attrs_key):
 
-            key = key.replace(zarr_attrs_key, self._attrs_key)
-            value = attrs_to_zarr(self._load_n5_attrs(key))
+            key_new = key.replace(zarr_attrs_key, self._attrs_key)
+            value = attrs_to_zarr(self._load_n5_attrs(key_new))
 
             if len(value) == 0:
-                raise KeyError(key)
+                raise KeyError(key_new)
             else:
                 return json_dumps(value)
 
         elif is_chunk_key(key):
-            key = self._swap_separator(key)
+            key_new = self._swap_separator(key)
 
-        return super().__getitem__(key)
+        else:
+            key_new = key
 
-    def __setitem__(self, key, value):
+        return super().__getitem__(key_new)
+
+    def __setitem__(self, key: str, value: Any):
         if key.endswith(zarr_group_meta_key):
 
-            key = key.replace(zarr_group_meta_key, self._group_meta_key)
+            key_new = key.replace(zarr_group_meta_key, self._group_meta_key)
 
-            n5_attrs = self._load_n5_attrs(key)
+            n5_attrs = self._load_n5_attrs(key_new)
             n5_attrs.update(**group_metadata_to_n5(json_loads(value)))
 
             value = json_dumps(n5_attrs)
 
         elif key.endswith(zarr_array_meta_key):
 
-            key = key.replace(zarr_array_meta_key, self._array_meta_key)
-
-            n5_attrs = self._load_n5_attrs(key)
-            n5_attrs.update(**array_metadata_to_n5(json_loads(value)))
+            key_new = key.replace(zarr_array_meta_key, self._array_meta_key)
+            top_level = key == zarr_array_meta_key
+            n5_attrs = self._load_n5_attrs(key_new)
+            n5_attrs.update(**array_metadata_to_n5(json_loads(value), top_level=top_level))
 
             value = json_dumps(n5_attrs)
 
         elif key.endswith(zarr_attrs_key):
 
-            key = key.replace(zarr_attrs_key, self._attrs_key)
+            key_new = key.replace(zarr_attrs_key, self._attrs_key)
 
-            n5_attrs = self._load_n5_attrs(key)
+            n5_attrs = self._load_n5_attrs(key_new)
             zarr_attrs = json_loads(value)
 
             for k in n5_keywords:
                 if k in zarr_attrs.keys():
-                    warnings.warn("attribute %s is a reserved N5 keyword" % k, UserWarning)
+                    warnings.warn(f"Attribute {k} is a reserved N5 keyword", UserWarning)
 
             # replace previous user attributes
             for k in list(n5_attrs.keys()):
@@ -437,52 +450,58 @@ class N5FSStore(FSStore):
             value = json_dumps(n5_attrs)
 
         elif is_chunk_key(key):
-            key = self._swap_separator(key)
+            key_new = self._swap_separator(key)
 
-        super().__setitem__(key, value)
+        else:
+            key_new = key
 
-    def __delitem__(self, key):
+        super().__setitem__(key_new, value)
 
-        if key.endswith(zarr_group_meta_key):  # pragma: no cover
-            key = key.replace(zarr_group_meta_key, self._group_meta_key)
-        elif key.endswith(zarr_array_meta_key):  # pragma: no cover
-            key = key.replace(zarr_array_meta_key, self._array_meta_key)
-        elif key.endswith(zarr_attrs_key):  # pragma: no cover
-            key = key.replace(zarr_attrs_key, self._attrs_key)
+    def __delitem__(self, key: str):
+
+        if key.endswith(zarr_group_meta_key):
+            key_new = key.replace(zarr_group_meta_key, self._group_meta_key)
+        elif key.endswith(zarr_array_meta_key):
+            key_new = key.replace(zarr_array_meta_key, self._array_meta_key)
+        elif key.endswith(zarr_attrs_key):
+            key_new = key.replace(zarr_attrs_key, self._attrs_key)
         elif is_chunk_key(key):
-            key = self._swap_separator(key)
+            key_new = self._swap_separator(key)
+        else:
+            key_new = key
+        super().__delitem__(key_new)
 
-        super().__delitem__(key)
-
-    def __contains__(self, key):
+    def __contains__(self, key: Any):
         if key.endswith(zarr_group_meta_key):
 
-            key = key.replace(zarr_group_meta_key, self._group_meta_key)
-            if key not in self:
+            key_new = key.replace(zarr_group_meta_key, self._group_meta_key)
+            if key_new not in self:
                 return False
             # group if not a dataset (attributes do not contain 'dimensions')
-            return "dimensions" not in self._load_n5_attrs(key)
+            return "dimensions" not in self._load_n5_attrs(key_new)
 
         elif key.endswith(zarr_array_meta_key):
 
-            key = key.replace(zarr_array_meta_key, self._array_meta_key)
+            key_new = key.replace(zarr_array_meta_key, self._array_meta_key)
             # array if attributes contain 'dimensions'
-            return "dimensions" in self._load_n5_attrs(key)
+            return "dimensions" in self._load_n5_attrs(key_new)
 
         elif key.endswith(zarr_attrs_key):
 
-            key = key.replace(zarr_attrs_key, self._attrs_key)
-            return self._contains_attrs(key)
+            key_new = key.replace(zarr_attrs_key, self._attrs_key)
+            return self._contains_attrs(key_new)
 
         elif is_chunk_key(key):
-            key = self._swap_separator(key)
+            key_new = self._swap_separator(key)
 
-        return super().__contains__(key)
+        else:
+            key_new = key
+        return super().__contains__(key_new)
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any):
         return isinstance(other, N5FSStore) and self.path == other.path
 
-    def listdir(self, path=None):
+    def listdir(self, path: Optional[str] = None):
         if path is not None:
             path = invert_chunk_coords(path)
 
@@ -519,20 +538,20 @@ class N5FSStore(FSStore):
             # replace n5 attribute file with respective zarr attribute files
             children.remove(self._group_meta_key)
             children.append(zarr_group_meta_key)
-            if self._contains_attrs(path):  # pragma: no cover
+            if self._contains_attrs(path):
                 children.append(zarr_attrs_key)
             return sorted(children)
         else:
             return children
 
-    def _load_n5_attrs(self, path):
+    def _load_n5_attrs(self, path: str):
         try:
             s = super().__getitem__(path)
             return json_loads(s)
         except KeyError:
             return {}
 
-    def _is_group(self, path):
+    def _is_group(self, path: Optional[str]):
 
         if path is None:
             attrs_key = self._attrs_key
@@ -542,7 +561,7 @@ class N5FSStore(FSStore):
         n5_attrs = self._load_n5_attrs(attrs_key)
         return len(n5_attrs) > 0 and "dimensions" not in n5_attrs
 
-    def _is_array(self, path):
+    def _is_array(self, path: Optional[str]):
 
         if path is None:
             attrs_key = self._attrs_key
@@ -551,30 +570,30 @@ class N5FSStore(FSStore):
 
         return "dimensions" in self._load_n5_attrs(attrs_key)
 
-    def _contains_attrs(self, path):
+    def _contains_attrs(self, path: Optional[str]):
 
         if path is None:
             attrs_key = self._attrs_key
         else:
             if not path.endswith(self._attrs_key):
                 attrs_key = os.path.join(path, self._attrs_key)
-            else:  # pragma: no cover
+            else:
                 attrs_key = path
 
         attrs = attrs_to_zarr(self._load_n5_attrs(attrs_key))
         return len(attrs) > 0
 
 
-def is_chunk_key(key):
+def is_chunk_key(key: str):
     rv = False
     segments = list(key.split('/'))
     if segments:
         last_segment = segments[-1]
-        rv = _prog_ckey.match(last_segment)
+        rv = bool(_prog_ckey.match(last_segment))
     return rv
 
 
-def invert_chunk_coords(key):
+def invert_chunk_coords(key: str):
     segments = list(key.split('/'))
     if segments:
         last_segment = segments[-1]
@@ -586,15 +605,15 @@ def invert_chunk_coords(key):
     return key
 
 
-def group_metadata_to_n5(group_metadata):
+def group_metadata_to_n5(group_metadata: Dict[str, Any]) -> Dict[str, Any]:
     '''Convert group metadata from zarr to N5 format.'''
     del group_metadata['zarr_format']
     # TODO: This should only exist at the top-level
-    group_metadata['n5'] = '2.0.0'
+    group_metadata['n5'] = N5_FORMAT
     return group_metadata
 
 
-def group_metadata_to_zarr(group_metadata):
+def group_metadata_to_zarr(group_metadata: Dict[str, Any]) -> Dict[str, Any]:
     '''Convert group metadata from N5 to zarr format.'''
     # This only exists at the top level
     group_metadata.pop('n5', None)
@@ -602,19 +621,20 @@ def group_metadata_to_zarr(group_metadata):
     return group_metadata
 
 
-def array_metadata_to_n5(array_metadata):
-    '''Convert array metadata from zarr to N5 format.'''
+def array_metadata_to_n5(array_metadata: Dict[str, Any], top_level=False) -> Dict[str, Any]:
+    '''Convert array metadata from zarr to N5 format. If the `top_level` keyword argument is True,
+    then the `N5` : N5_FORMAT key : value pair will be inserted into the metadata.'''
 
     for f, t in zarr_to_n5_keys:
-        array_metadata[t] = array_metadata[f]
-        del array_metadata[f]
+        array_metadata[t] = array_metadata.pop(f)
     del array_metadata['zarr_format']
-
+    if top_level:
+        array_metadata['n5'] = N5_FORMAT
     try:
         dtype = np.dtype(array_metadata['dataType'])
-    except TypeError:  # pragma: no cover
+    except TypeError:
         raise TypeError(
-            "data type %s not supported by N5" % array_metadata['dataType'])
+            f"Data type {array_metadata['dataType']} is not supported by N5")
 
     array_metadata['dataType'] = dtype.name
     array_metadata['dimensions'] = array_metadata['dimensions'][::-1]
@@ -622,17 +642,24 @@ def array_metadata_to_n5(array_metadata):
 
     if 'fill_value' in array_metadata:
         if array_metadata['fill_value'] != 0 and array_metadata['fill_value'] is not None:
-            raise ValueError("N5 only supports fill_value == 0 (for now)")
+            raise ValueError(
+                f'''Received fill_value = {array_metadata['fill_value']},
+                but N5 only supports fill_value = 0'''
+                )
         del array_metadata['fill_value']
 
     if 'order' in array_metadata:
         if array_metadata['order'] != 'C':
-            raise ValueError("zarr N5 storage only stores arrays in C order (for now)")
+            raise ValueError(
+                f"Received order = {array_metadata['order']}, but N5 only supports order = C"
+                )
         del array_metadata['order']
 
     if 'filters' in array_metadata:
         if array_metadata['filters'] != [] and array_metadata['filters'] is not None:
-            raise ValueError("N5 storage does not support zarr filters")
+            raise ValueError(
+                "Received filters, but N5 storage does not support zarr filters"
+                )
         del array_metadata['filters']
 
     assert 'compression' in array_metadata
@@ -646,11 +673,14 @@ def array_metadata_to_n5(array_metadata):
     return array_metadata
 
 
-def array_metadata_to_zarr(array_metadata):
-    '''Convert array metadata from N5 to zarr format.'''
+def array_metadata_to_zarr(array_metadata: Dict[str, Any],
+                           top_level: bool = False) -> Dict[str, Any]:
+    '''Convert array metadata from N5 to zarr format.
+    If the `top_level` keyword argument is True, then the `N5` key will be removed from metadata'''
     for t, f in zarr_to_n5_keys:
-        array_metadata[t] = array_metadata[f]
-        del array_metadata[f]
+        array_metadata[t] = array_metadata.pop(f)
+    if top_level:
+        array_metadata.pop('n5')
     array_metadata['zarr_format'] = ZARR_FORMAT
 
     array_metadata['shape'] = array_metadata['shape'][::-1]
@@ -672,7 +702,7 @@ def array_metadata_to_zarr(array_metadata):
     return array_metadata
 
 
-def attrs_to_zarr(attrs):
+def attrs_to_zarr(attrs: Dict[str, Any]) -> Dict[str, Any]:
     '''Get all zarr attributes from an N5 attributes dictionary (i.e.,
     all non-keyword attributes).'''
 
@@ -684,22 +714,24 @@ def attrs_to_zarr(attrs):
     return attrs
 
 
-def compressor_config_to_n5(compressor_config):
+def compressor_config_to_n5(compressor_config: Optional[Dict[str, Any]]) -> Dict[str, Any]:
 
     if compressor_config is None:
         return {'type': 'raw'}
+    else:
+        _compressor_config = compressor_config
 
     # peel wrapper, if present
-    if compressor_config['id'] == N5ChunkWrapper.codec_id:
-        compressor_config = compressor_config['compressor_config']
+    if _compressor_config['id'] == N5ChunkWrapper.codec_id:
+        _compressor_config = _compressor_config['compressor_config']
 
-    codec_id = compressor_config['id']
+    codec_id = _compressor_config['id']
     n5_config = {'type': codec_id}
 
     if codec_id == 'bz2':
 
         n5_config['type'] = 'bzip2'
-        n5_config['blockSize'] = compressor_config['level']
+        n5_config['blockSize'] = _compressor_config['level']
 
     elif codec_id == 'blosc':
 
@@ -709,16 +741,16 @@ def compressor_config_to_n5(compressor_config):
             RuntimeWarning
         )
 
-        n5_config['cname'] = compressor_config['cname']
-        n5_config['clevel'] = compressor_config['clevel']
-        n5_config['shuffle'] = compressor_config['shuffle']
-        n5_config['blocksize'] = compressor_config['blocksize']
+        n5_config['cname'] = _compressor_config['cname']
+        n5_config['clevel'] = _compressor_config['clevel']
+        n5_config['shuffle'] = _compressor_config['shuffle']
+        n5_config['blocksize'] = _compressor_config['blocksize']
 
     elif codec_id == 'lzma':
 
         # Switch to XZ for N5 if we are using the default XZ format.
         # Note: 4 is the default, which is lzma.CHECK_CRC64.
-        if compressor_config['format'] == 1 and compressor_config['check'] in [-1, 4]:
+        if _compressor_config['format'] == 1 and _compressor_config['check'] in [-1, 4]:
             n5_config['type'] = 'xz'
         else:
             warnings.warn(
@@ -726,36 +758,36 @@ def compressor_config_to_n5(compressor_config):
                 "might not be able to open the dataset with another N5 library.",
                 RuntimeWarning
             )
-            n5_config['format'] = compressor_config['format']
-            n5_config['check'] = compressor_config['check']
-            n5_config['filters'] = compressor_config['filters']
+            n5_config['format'] = _compressor_config['format']
+            n5_config['check'] = _compressor_config['check']
+            n5_config['filters'] = _compressor_config['filters']
 
         # The default is lzma.PRESET_DEFAULT, which is 6.
-        if compressor_config['preset']:
-            n5_config['preset'] = compressor_config['preset']
+        if _compressor_config['preset']:
+            n5_config['preset'] = _compressor_config['preset']
         else:
             n5_config['preset'] = 6
 
     elif codec_id == 'zlib':
 
         n5_config['type'] = 'gzip'
-        n5_config['level'] = compressor_config['level']
+        n5_config['level'] = _compressor_config['level']
         n5_config['useZlib'] = True
 
-    elif codec_id == 'gzip':  # pragma: no cover
+    elif codec_id == 'gzip':
 
         n5_config['type'] = 'gzip'
-        n5_config['level'] = compressor_config['level']
+        n5_config['level'] = _compressor_config['level']
         n5_config['useZlib'] = False
 
-    else:  # pragma: no cover
+    else:
 
-        n5_config.update({k: v for k, v in compressor_config.items() if k != 'type'})
+        n5_config.update({k: v for k, v in _compressor_config.items() if k != 'type'})
 
     return n5_config
 
 
-def compressor_config_to_zarr(compressor_config):
+def compressor_config_to_zarr(compressor_config: Dict[str, Any]) -> Optional[Dict[str, Any]]:
 
     codec_id = compressor_config['type']
     zarr_config = {'id': codec_id}
@@ -792,7 +824,7 @@ def compressor_config_to_zarr(compressor_config):
         if 'useZlib' in compressor_config and compressor_config['useZlib']:
             zarr_config['id'] = 'zlib'
             zarr_config['level'] = compressor_config['level']
-        else:  # pragma: no cover
+        else:
             zarr_config['id'] = 'gzip'
             zarr_config['level'] = compressor_config['level']
 
@@ -800,7 +832,7 @@ def compressor_config_to_zarr(compressor_config):
 
         return None
 
-    else:  # pragma: no cover
+    else:
 
         zarr_config.update({k: v for k, v in compressor_config.items() if k != 'type'})
 
@@ -821,7 +853,7 @@ class N5ChunkWrapper(Codec):
             (self.dtype.byteorder == '=' and sys.byteorder == 'little')
         )
 
-        if compressor:  # pragma: no cover
+        if compressor:
             if compressor_config is not None:
                 raise ValueError("Only one of compressor_config or compressor should be given.")
             compressor_config = compressor.get_config()
@@ -854,7 +886,7 @@ class N5ChunkWrapper(Codec):
         else:
             return header + chunk.tobytes(order='A')
 
-    def decode(self, chunk, out=None):
+    def decode(self, chunk, out=None) -> bytes:
 
         len_header, chunk_shape = self._read_header(chunk)
         chunk = chunk[len_header:]
@@ -887,7 +919,7 @@ class N5ChunkWrapper(Codec):
             chunk = self._from_big_endian(chunk)
 
             # read partial chunk
-            if chunk_shape != self.chunk_shape:  # pragma: no cover
+            if chunk_shape != self.chunk_shape:
                 chunk = np.frombuffer(chunk, dtype=self.dtype)
                 chunk = chunk.reshape(chunk_shape)
                 complete_chunk = np.zeros(self.chunk_shape, dtype=self.dtype)
@@ -897,7 +929,8 @@ class N5ChunkWrapper(Codec):
 
             return chunk
 
-    def _create_header(self, chunk):
+    @staticmethod
+    def _create_header(chunk):
 
         mode = struct.pack('>H', 0)
         num_dims = struct.pack('>H', len(chunk.shape))
@@ -908,7 +941,8 @@ class N5ChunkWrapper(Codec):
 
         return mode + num_dims + shape
 
-    def _read_header(self, chunk):
+    @staticmethod
+    def _read_header(chunk):
 
         num_dims = struct.unpack('>H', chunk[2:4])[0]
         shape = tuple(
