@@ -24,28 +24,57 @@ from zarr.convenience import (
 from zarr.core import Array
 from zarr.errors import CopyError
 from zarr.hierarchy import Group, group
-from zarr.storage import (ConsolidatedMetadataStore, MemoryStore,
-                          atexit_rmtree, getsize)
+from zarr.storage import (
+    ConsolidatedMetadataStore,
+    FSStore,
+    KVStore,
+    MemoryStore,
+    atexit_rmtree,
+    data_root,
+    meta_root,
+    getsize,
+)
+from zarr._storage.store import v3_api_available
+from zarr._storage.v3 import (
+    ConsolidatedMetadataStoreV3,
+    DirectoryStoreV3,
+    FSStoreV3,
+    KVStoreV3,
+    MemoryStoreV3,
+    SQLiteStoreV3,
+)
+from zarr.tests.util import have_fsspec
+
+_VERSIONS = ((2, 3) if v3_api_available else (2, ))
 
 
-def test_open_array(path_type):
+def _init_creation_kwargs(zarr_version):
+    kwargs = {'zarr_version': zarr_version}
+    if zarr_version == 3:
+        kwargs['path'] = 'dataset'
+    return kwargs
+
+
+@pytest.mark.parametrize('zarr_version', _VERSIONS)
+def test_open_array(path_type, zarr_version):
 
     store = tempfile.mkdtemp()
     atexit.register(atexit_rmtree, store)
     store = path_type(store)
+    kwargs = _init_creation_kwargs(zarr_version)
 
     # open array, create if doesn't exist
-    z = open(store, mode='a', shape=100)
+    z = open(store, mode='a', shape=100, **kwargs)
     assert isinstance(z, Array)
     assert z.shape == (100,)
 
     # open array, overwrite
-    z = open(store, mode='w', shape=200)
+    z = open(store, mode='w', shape=200, **kwargs)
     assert isinstance(z, Array)
     assert z.shape == (200,)
 
     # open array, read-only
-    z = open(store, mode='r')
+    z = open(store, mode='r', **kwargs)
     assert isinstance(z, Array)
     assert z.shape == (200,)
     assert z.read_only
@@ -55,44 +84,80 @@ def test_open_array(path_type):
         open('doesnotexist', mode='r')
 
 
-def test_open_group(path_type):
+@pytest.mark.parametrize("zarr_version", _VERSIONS)
+def test_open_group(path_type, zarr_version):
 
     store = tempfile.mkdtemp()
     atexit.register(atexit_rmtree, store)
     store = path_type(store)
+    kwargs = _init_creation_kwargs(zarr_version)
 
     # open group, create if doesn't exist
-    g = open(store, mode='a')
+    g = open(store, mode='a', **kwargs)
     g.create_group('foo')
     assert isinstance(g, Group)
     assert 'foo' in g
 
     # open group, overwrite
-    g = open(store, mode='w')
+    g = open(store, mode='w', **kwargs)
     assert isinstance(g, Group)
     assert 'foo' not in g
 
     # open group, read-only
-    g = open(store, mode='r')
+    g = open(store, mode='r', **kwargs)
     assert isinstance(g, Group)
     assert g.read_only
 
 
-def test_save_errors():
+@pytest.mark.parametrize("zarr_version", _VERSIONS)
+def test_save_errors(zarr_version):
     with pytest.raises(ValueError):
         # no arrays provided
-        save_group('data/group.zarr')
+        save_group('data/group.zarr', zarr_version=zarr_version)
+    with pytest.raises(TypeError):
+        # no array provided
+        save_array('data/group.zarr', zarr_version=zarr_version)
     with pytest.raises(ValueError):
         # no arrays provided
-        save('data/group.zarr')
+        save('data/group.zarr', zarr_version=zarr_version)
 
 
-def test_lazy_loader():
+@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
+def test_zarr_v3_save_multiple_unnamed():
+    x = np.ones(8)
+    y = np.zeros(8)
+    store = KVStoreV3(dict())
+    # no path provided
+    save_group(store, x, y, path='dataset', zarr_version=3)
+    # names become arr_{i} for unnamed *args
+    assert data_root + 'dataset/arr_0/c0' in store
+    assert data_root + 'dataset/arr_1/c0' in store
+    assert meta_root + 'dataset/arr_0.array.json' in store
+    assert meta_root + 'dataset/arr_1.array.json' in store
+
+
+@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
+def test_zarr_v3_save_errors():
+    x = np.ones(8)
+    with pytest.raises(ValueError):
+        # no path provided
+        save_group('data/group.zr3', x, zarr_version=3)
+    with pytest.raises(ValueError):
+        # no path provided
+        save_array('data/group.zr3', x, zarr_version=3)
+    with pytest.raises(ValueError):
+        # no path provided
+        save('data/group.zr3', x, zarr_version=3)
+
+
+@pytest.mark.parametrize("zarr_version", _VERSIONS)
+def test_lazy_loader(zarr_version):
     foo = np.arange(100)
     bar = np.arange(100, 0, -1)
-    store = 'data/group.zarr'
-    save(store, foo=foo, bar=bar)
-    loader = load(store)
+    store = 'data/group.zarr' if zarr_version == 2 else 'data/group.zr3'
+    kwargs = _init_creation_kwargs(zarr_version)
+    save(store, foo=foo, bar=bar, **kwargs)
+    loader = load(store, **kwargs)
     assert 'foo' in loader
     assert 'bar' in loader
     assert 'baz' not in loader
@@ -100,13 +165,78 @@ def test_lazy_loader():
     assert sorted(loader) == ['bar', 'foo']
     assert_array_equal(foo, loader['foo'])
     assert_array_equal(bar, loader['bar'])
+    assert 'LazyLoader: ' in repr(loader)
 
 
-def test_consolidate_metadata():
+@pytest.mark.parametrize("zarr_version", _VERSIONS)
+def test_load_array(zarr_version):
+    foo = np.arange(100)
+    bar = np.arange(100, 0, -1)
+    store = 'data/group.zarr' if zarr_version == 2 else 'data/group.zr3'
+    kwargs = _init_creation_kwargs(zarr_version)
+    save(store, foo=foo, bar=bar, **kwargs)
+
+    # can also load arrays directly into a numpy array
+    for array_name in ['foo', 'bar']:
+        array_path = 'dataset/' + array_name if zarr_version == 3 else array_name
+        array = load(store, path=array_path, zarr_version=zarr_version)
+        assert isinstance(array, np.ndarray)
+        if array_name == 'foo':
+            assert_array_equal(foo, array)
+        else:
+            assert_array_equal(bar, array)
+
+
+@pytest.mark.parametrize("zarr_version", _VERSIONS)
+def test_tree(zarr_version):
+    kwargs = _init_creation_kwargs(zarr_version)
+    g1 = zarr.group(**kwargs)
+    g1.create_group('foo')
+    g3 = g1.create_group('bar')
+    g3.create_group('baz')
+    g5 = g3.create_group('qux')
+    g5.create_dataset('baz', shape=100, chunks=10)
+    assert repr(zarr.tree(g1)) == repr(g1.tree())
+    assert str(zarr.tree(g1)) == str(g1.tree())
+
+
+@pytest.mark.parametrize('zarr_version', _VERSIONS)
+@pytest.mark.parametrize('stores_from_path', [False, True])
+@pytest.mark.parametrize(
+    'with_chunk_store,listable',
+    [(False, True), (True, True), (False, False)],
+    ids=['default-listable', 'with_chunk_store-listable', 'default-unlistable']
+)
+def test_consolidate_metadata(with_chunk_store,
+                              zarr_version,
+                              listable,
+                              monkeypatch,
+                              stores_from_path):
 
     # setup initial data
-    store = MemoryStore()
-    z = group(store)
+    if stores_from_path:
+        store = tempfile.mkdtemp()
+        atexit.register(atexit_rmtree, store)
+        if with_chunk_store:
+            chunk_store = tempfile.mkdtemp()
+            atexit.register(atexit_rmtree, chunk_store)
+        else:
+            chunk_store = None
+        version_kwarg = {'zarr_version': zarr_version}
+    else:
+        if zarr_version == 2:
+            store = MemoryStore()
+            chunk_store = MemoryStore() if with_chunk_store else None
+        elif zarr_version == 3:
+            store = MemoryStoreV3()
+            chunk_store = MemoryStoreV3() if with_chunk_store else None
+        version_kwarg = {}
+    path = 'dataset' if zarr_version == 3 else None
+    z = group(store, chunk_store=chunk_store, path=path, **version_kwarg)
+
+    # Reload the actual store implementation in case str
+    store_to_copy = z.store
+
     z.create_group('g1')
     g2 = z.create_group('g2')
     g2.attrs['hello'] = 'world'
@@ -117,36 +247,103 @@ def test_consolidate_metadata():
     arr[:] = 1.0
     assert 16 == arr.nchunks_initialized
 
+    if stores_from_path:
+        # get the actual store class for use with consolidate_metadata
+        store_class = z._store
+    else:
+        store_class = store
+
+    if zarr_version == 3:
+        # error on v3 if path not provided
+        with pytest.raises(ValueError):
+            consolidate_metadata(store_class, path=None)
+
+        with pytest.raises(ValueError):
+            consolidate_metadata(store_class, path='')
+
     # perform consolidation
-    out = consolidate_metadata(store)
+    out = consolidate_metadata(store_class, path=path)
     assert isinstance(out, Group)
-    assert '.zmetadata' in store
-    for key in ['.zgroup',
-                'g1/.zgroup',
-                'g2/.zgroup',
-                'g2/.zattrs',
-                'g2/arr/.zarray',
-                'g2/arr/.zattrs']:
-        del store[key]
+    assert ['g1', 'g2'] == list(out)
+    if not stores_from_path:
+        if zarr_version == 2:
+            assert isinstance(out._store, ConsolidatedMetadataStore)
+            assert '.zmetadata' in store
+            meta_keys = ['.zgroup',
+                         'g1/.zgroup',
+                         'g2/.zgroup',
+                         'g2/.zattrs',
+                         'g2/arr/.zarray',
+                         'g2/arr/.zattrs']
+        else:
+            assert isinstance(out._store, ConsolidatedMetadataStoreV3)
+            assert 'meta/root/consolidated/.zmetadata' in store
+            meta_keys = ['zarr.json',
+                         meta_root + 'dataset.group.json',
+                         meta_root + 'dataset/g1.group.json',
+                         meta_root + 'dataset/g2.group.json',
+                         meta_root + 'dataset/g2/arr.array.json',
+                         'meta/root/consolidated.group.json']
+        for key in meta_keys:
+            del store[key]
+
+    # https://github.com/zarr-developers/zarr-python/issues/993
+    # Make sure we can still open consolidated on an unlistable store:
+    if not listable:
+        fs_memory = pytest.importorskip("fsspec.implementations.memory")
+        monkeypatch.setattr(fs_memory.MemoryFileSystem, "isdir", lambda x, y: False)
+        monkeypatch.delattr(fs_memory.MemoryFileSystem, "ls")
+        fs = fs_memory.MemoryFileSystem()
+        if zarr_version == 2:
+            store_to_open = FSStore("", fs=fs)
+        else:
+            store_to_open = FSStoreV3("", fs=fs)
+
+        # copy original store to new unlistable store
+        store_to_open.update(store_to_copy)
+
+    else:
+        store_to_open = store
 
     # open consolidated
-    z2 = open_consolidated(store)
+    z2 = open_consolidated(store_to_open, chunk_store=chunk_store, path=path, **version_kwarg)
     assert ['g1', 'g2'] == list(z2)
     assert 'world' == z2.g2.attrs['hello']
     assert 1 == z2.g2.arr.attrs['data']
     assert (z2.g2.arr[:] == 1.0).all()
     assert 16 == z2.g2.arr.nchunks
-    assert 16 == z2.g2.arr.nchunks_initialized
+    if listable:
+        assert 16 == z2.g2.arr.nchunks_initialized
+    else:
+        with pytest.raises(NotImplementedError):
+            _ = z2.g2.arr.nchunks_initialized
 
-    # tests del/write on the store
-    cmd = ConsolidatedMetadataStore(store)
-    with pytest.raises(PermissionError):
-        del cmd['.zgroup']
-    with pytest.raises(PermissionError):
-        cmd['.zgroup'] = None
+    if stores_from_path:
+        # path string is note a BaseStore subclass so cannot be used to
+        # initialize a ConsolidatedMetadataStore.
+        if zarr_version == 2:
+            with pytest.raises(ValueError):
+                cmd = ConsolidatedMetadataStore(store)
+        elif zarr_version == 3:
+            with pytest.raises(ValueError):
+                cmd = ConsolidatedMetadataStoreV3(store)
+    else:
+        # tests del/write on the store
+        if zarr_version == 2:
+            cmd = ConsolidatedMetadataStore(store)
+            with pytest.raises(PermissionError):
+                del cmd['.zgroup']
+            with pytest.raises(PermissionError):
+                cmd['.zgroup'] = None
+        else:
+            cmd = ConsolidatedMetadataStoreV3(store)
+            with pytest.raises(PermissionError):
+                del cmd[meta_root + 'dataset.group.json']
+            with pytest.raises(PermissionError):
+                cmd[meta_root + 'dataset.group.json'] = None
 
-    # test getsize on the store
-    assert isinstance(getsize(cmd), Integral)
+        # test getsize on the store
+        assert isinstance(getsize(cmd), Integral)
 
     # test new metadata are not writeable
     with pytest.raises(PermissionError):
@@ -168,62 +365,17 @@ def test_consolidate_metadata():
 
     # test invalid modes
     with pytest.raises(ValueError):
-        open_consolidated(store, mode='a')
+        open_consolidated(store, chunk_store=chunk_store, mode='a', path=path)
     with pytest.raises(ValueError):
-        open_consolidated(store, mode='w')
+        open_consolidated(store, chunk_store=chunk_store, mode='w', path=path)
+    with pytest.raises(ValueError):
+        open_consolidated(store, chunk_store=chunk_store, mode='w-', path=path)
 
     # make sure keyword arguments are passed through without error
-    open_consolidated(store, cache_attrs=True, synchronizer=None)
-
-
-def test_consolidated_with_chunk_store():
-    # setup initial data
-    store = MemoryStore()
-    chunk_store = MemoryStore()
-    z = group(store, chunk_store=chunk_store)
-    z.create_group('g1')
-    g2 = z.create_group('g2')
-    g2.attrs['hello'] = 'world'
-    arr = g2.create_dataset('arr', shape=(20, 20), chunks=(5, 5), dtype='f8')
-    assert 16 == arr.nchunks
-    assert 0 == arr.nchunks_initialized
-    arr.attrs['data'] = 1
-    arr[:] = 1.0
-    assert 16 == arr.nchunks_initialized
-
-    # perform consolidation
-    out = consolidate_metadata(store)
-    assert isinstance(out, Group)
-    assert '.zmetadata' in store
-    for key in ['.zgroup',
-                'g1/.zgroup',
-                'g2/.zgroup',
-                'g2/.zattrs',
-                'g2/arr/.zarray',
-                'g2/arr/.zattrs']:
-        del store[key]
-    # open consolidated
-    z2 = open_consolidated(store, chunk_store=chunk_store)
-    assert ['g1', 'g2'] == list(z2)
-    assert 'world' == z2.g2.attrs['hello']
-    assert 1 == z2.g2.arr.attrs['data']
-    assert (z2.g2.arr[:] == 1.0).all()
-    assert 16 == z2.g2.arr.nchunks
-    assert 16 == z2.g2.arr.nchunks_initialized
-
-    # test the data are writeable
-    z2.g2.arr[:] = 2
-    assert (z2.g2.arr[:] == 2).all()
-
-    # test invalid modes
-    with pytest.raises(ValueError):
-        open_consolidated(store, mode='a', chunk_store=chunk_store)
-    with pytest.raises(ValueError):
-        open_consolidated(store, mode='w', chunk_store=chunk_store)
-
-    # make sure keyword arguments are passed through without error
-    open_consolidated(store, cache_attrs=True, synchronizer=None,
-                      chunk_store=chunk_store)
+    open_consolidated(
+        store, chunk_store=chunk_store, path=path, cache_attrs=True, synchronizer=None,
+        **version_kwarg,
+    )
 
 
 @pytest.mark.parametrize("options", (
@@ -239,6 +391,8 @@ def test_save_array_separator(tmpdir, options):
 
 class TestCopyStore(unittest.TestCase):
 
+    _version = 2
+
     def setUp(self):
         source = dict()
         source['foo'] = b'xxx'
@@ -246,9 +400,12 @@ class TestCopyStore(unittest.TestCase):
         source['bar/qux'] = b'zzz'
         self.source = source
 
+    def _get_dest_store(self):
+        return dict()
+
     def test_no_paths(self):
         source = self.source
-        dest = dict()
+        dest = self._get_dest_store()
         copy_store(source, dest)
         assert len(source) == len(dest)
         for key in source:
@@ -258,7 +415,7 @@ class TestCopyStore(unittest.TestCase):
         source = self.source
         # paths should be normalized
         for source_path in 'bar', 'bar/', '/bar', '/bar/':
-            dest = dict()
+            dest = self._get_dest_store()
             copy_store(source, dest, source_path=source_path)
             assert 2 == len(dest)
             for key in source:
@@ -272,11 +429,14 @@ class TestCopyStore(unittest.TestCase):
         source = self.source
         # paths should be normalized
         for dest_path in 'new', 'new/', '/new', '/new/':
-            dest = dict()
+            dest = self._get_dest_store()
             copy_store(source, dest, dest_path=dest_path)
             assert len(source) == len(dest)
             for key in source:
-                dest_key = 'new/' + key
+                if self._version == 3:
+                    dest_key = key[:10] + 'new/' + key[10:]
+                else:
+                    dest_key = 'new/' + key
                 assert source[key] == dest[dest_key]
 
     def test_source_dest_path(self):
@@ -284,7 +444,7 @@ class TestCopyStore(unittest.TestCase):
         # paths should be normalized
         for source_path in 'bar', 'bar/', '/bar', '/bar/':
             for dest_path in 'new', 'new/', '/new', '/new/':
-                dest = dict()
+                dest = self._get_dest_store()
                 copy_store(source, dest, source_path=source_path,
                            dest_path=dest_path)
                 assert 2 == len(dest)
@@ -300,41 +460,44 @@ class TestCopyStore(unittest.TestCase):
         source = self.source
 
         # single excludes
-        dest = dict()
+        dest = self._get_dest_store()
         excludes = 'f.*'
         copy_store(source, dest, excludes=excludes)
         assert len(dest) == 2
-        assert 'foo' not in dest
+
+        root = '' if self._version == 2 else meta_root
+        assert root + 'foo' not in dest
 
         # multiple excludes
-        dest = dict()
+        dest = self._get_dest_store()
         excludes = 'b.z', '.*x'
         copy_store(source, dest, excludes=excludes)
         assert len(dest) == 1
-        assert 'foo' in dest
-        assert 'bar/baz' not in dest
-        assert 'bar/qux' not in dest
+        assert root + 'foo' in dest
+        assert root + 'bar/baz' not in dest
+        assert root + 'bar/qux' not in dest
 
         # excludes and includes
-        dest = dict()
+        dest = self._get_dest_store()
         excludes = 'b.*'
         includes = '.*x'
         copy_store(source, dest, excludes=excludes, includes=includes)
         assert len(dest) == 2
-        assert 'foo' in dest
-        assert 'bar/baz' not in dest
-        assert 'bar/qux' in dest
+        assert root + 'foo' in dest
+        assert root + 'bar/baz' not in dest
+        assert root + 'bar/qux' in dest
 
     def test_dry_run(self):
         source = self.source
-        dest = dict()
+        dest = self._get_dest_store()
         copy_store(source, dest, dry_run=True)
         assert 0 == len(dest)
 
     def test_if_exists(self):
         source = self.source
-        dest = dict()
-        dest['bar/baz'] = b'mmm'
+        dest = self._get_dest_store()
+        root = '' if self._version == 2 else meta_root
+        dest[root + 'bar/baz'] = b'mmm'
 
         # default ('raise')
         with pytest.raises(CopyError):
@@ -347,22 +510,44 @@ class TestCopyStore(unittest.TestCase):
         # skip
         copy_store(source, dest, if_exists='skip')
         assert 3 == len(dest)
-        assert dest['foo'] == b'xxx'
-        assert dest['bar/baz'] == b'mmm'
-        assert dest['bar/qux'] == b'zzz'
+        assert dest[root + 'foo'] == b'xxx'
+        assert dest[root + 'bar/baz'] == b'mmm'
+        assert dest[root + 'bar/qux'] == b'zzz'
 
         # replace
         copy_store(source, dest, if_exists='replace')
         assert 3 == len(dest)
-        assert dest['foo'] == b'xxx'
-        assert dest['bar/baz'] == b'yyy'
-        assert dest['bar/qux'] == b'zzz'
+        assert dest[root + 'foo'] == b'xxx'
+        assert dest[root + 'bar/baz'] == b'yyy'
+        assert dest[root + 'bar/qux'] == b'zzz'
 
         # invalid option
         with pytest.raises(ValueError):
             copy_store(source, dest, if_exists='foobar')
 
 
+@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
+class TestCopyStoreV3(TestCopyStore):
+
+    _version = 3
+
+    def setUp(self):
+        source = KVStoreV3(dict())
+        source['meta/root/foo'] = b'xxx'
+        source['meta/root/bar/baz'] = b'yyy'
+        source['meta/root/bar/qux'] = b'zzz'
+        self.source = source
+
+    def _get_dest_store(self):
+        return KVStoreV3(dict())
+
+    def test_mismatched_store_versions(self):
+        # cannot copy between stores of mixed Zarr versions
+        dest = KVStore(dict())
+        with pytest.raises(ValueError):
+            copy_store(self.source, dest)
+
+
 def check_copied_array(original, copied, without_attrs=False,
                        expect_props=None):
 
@@ -411,7 +596,14 @@ def check_copied_array(original, copied, without_attrs=False,
         for k in original.attrs.keys():
             assert k not in copied.attrs
     else:
-        assert sorted(original.attrs.items()) == sorted(copied.attrs.items())
+        if dest_h5py and 'filters' in original.attrs:
+            # special case in v3 (storing filters metadata under attributes)
+            # we explicitly do not copy this info over to HDF5
+            original_attrs = original.attrs.asdict().copy()
+            original_attrs.pop('filters')
+        else:
+            original_attrs = original.attrs
+        assert sorted(original_attrs.items()) == sorted(copied.attrs.items())
 
 
 def check_copied_group(original, copied, without_attrs=False, expect_props=None,
@@ -465,10 +657,33 @@ def test_copy_all():
         dry_run=False,
     )
 
+    assert 'subgroup' in destination_group
     assert destination_group.attrs["info"] == "group attrs"
     assert destination_group.subgroup.attrs["info"] == "sub attrs"
 
 
+@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
+def test_copy_all_v3():
+    """
+    https://github.com/zarr-developers/zarr-python/issues/269
+
+    copy_all used to not copy attributes as `.keys()`
+
+    """
+    original_group = zarr.group(store=MemoryStoreV3(), path='group1', overwrite=True)
+    original_group.create_group("subgroup")
+
+    destination_group = zarr.group(store=MemoryStoreV3(), path='group2', overwrite=True)
+
+    # copy from memory to directory store
+    copy_all(
+        original_group,
+        destination_group,
+        dry_run=False,
+    )
+    assert 'subgroup' in destination_group
+
+
 class TestCopy:
     @pytest.fixture(params=[False, True], ids=['zarr', 'hdf5'])
     def source(self, request, tmpdir):
@@ -711,3 +926,89 @@ class TestCopy:
         # bad option
         with pytest.raises(TypeError):
             copy(source['foo'], dest, dry_run=True, log=True)
+
+
+@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
+class TestCopyV3(TestCopy):
+
+    @pytest.fixture(params=['zarr', 'hdf5'])
+    def source(self, request, tmpdir):
+        def prep_source(source):
+            foo = source.create_group('foo')
+            foo.attrs['experiment'] = 'weird science'
+            baz = foo.create_dataset('bar/baz', data=np.arange(100), chunks=(50,))
+            baz.attrs['units'] = 'metres'
+            if request.param == 'hdf5':
+                extra_kws = dict(compression='gzip', compression_opts=3, fillvalue=84,
+                                 shuffle=True, fletcher32=True)
+            else:
+                extra_kws = dict(compressor=Zlib(3), order='F', fill_value=42, filters=[Adler32()])
+            source.create_dataset('spam', data=np.arange(100, 200).reshape(20, 5),
+                                  chunks=(10, 2), dtype='i2', **extra_kws)
+            return source
+
+        if request.param == 'hdf5':
+            h5py = pytest.importorskip('h5py')
+            fn = tmpdir.join('source.h5')
+            with h5py.File(str(fn), mode='w') as h5f:
+                yield prep_source(h5f)
+        elif request.param == 'zarr':
+            yield prep_source(group(path='group1', zarr_version=3))
+
+    # Test with various destination StoreV3 types as TestCopyV3 covers rmdir
+    destinations = ['hdf5', 'zarr', 'zarr_kvstore', 'zarr_directorystore', 'zarr_sqlitestore']
+    if have_fsspec:
+        destinations += ['zarr_fsstore']
+
+    @pytest.fixture(params=destinations)
+    def dest(self, request, tmpdir):
+        if request.param == 'hdf5':
+            h5py = pytest.importorskip('h5py')
+            fn = tmpdir.join('dest.h5')
+            with h5py.File(str(fn), mode='w') as h5f:
+                yield h5f
+        elif request.param == 'zarr':
+            yield group(path='group2', zarr_version=3)
+        elif request.param == 'zarr_kvstore':
+            store = KVStoreV3(dict())
+            yield group(store, path='group2', zarr_version=3)
+        elif request.param == 'zarr_fsstore':
+            fn = tmpdir.join('dest.zr3')
+            store = FSStoreV3(str(fn), auto_mkdir=True)
+            yield group(store, path='group2', zarr_version=3)
+        elif request.param == 'zarr_directorystore':
+            fn = tmpdir.join('dest.zr3')
+            store = DirectoryStoreV3(str(fn))
+            yield group(store, path='group2', zarr_version=3)
+        elif request.param == 'zarr_sqlitestore':
+            fn = tmpdir.join('dest.db')
+            store = SQLiteStoreV3(str(fn))
+            yield group(store, path='group2', zarr_version=3)
+
+    def test_copy_array_create_options(self, source, dest):
+        dest_h5py = dest.__module__.startswith('h5py.')
+
+        # copy array, provide creation options
+        compressor = Zlib(9)
+        create_kws = dict(chunks=(10,))
+        if dest_h5py:
+            create_kws.update(compression='gzip', compression_opts=9,
+                              shuffle=True, fletcher32=True, fillvalue=42)
+        else:
+            # v3 case has no filters argument in zarr create_kws
+            create_kws.update(compressor=compressor, fill_value=42, order='F')
+        copy(source['foo/bar/baz'], dest, without_attrs=True, **create_kws)
+        check_copied_array(source['foo/bar/baz'], dest['baz'],
+                           without_attrs=True, expect_props=create_kws)
+
+    def test_copy_group_no_name(self, source, dest):
+        if source.__module__.startswith('h5py'):
+            with pytest.raises(TypeError):
+                copy(source, dest)
+        else:
+            # For v3, dest.name will be inferred from source.name
+            copy(source, dest)
+            check_copied_group(source, dest[source.name.lstrip('/')])
+
+        copy(source, dest, name='root')
+        check_copied_group(source, dest['root'])
